/*
 * This file is part of vospace-ui
 * Copyright (C) 2021 Istituto Nazionale di Astrofisica
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
package it.inaf.ia2.vospace.ui.controller;

import it.inaf.ia2.aa.data.User;
import it.inaf.ia2.vospace.ui.client.VOSpaceClient;
import it.inaf.ia2.vospace.ui.data.Job;
import it.inaf.ia2.vospace.ui.data.ListNodeData;
import it.inaf.ia2.vospace.ui.exception.VOSpaceException;
import it.inaf.ia2.vospace.ui.service.MainNodesHtmlGenerator;
import it.inaf.ia2.vospace.ui.service.MoveNodeModalHtmlGenerator;
import it.inaf.oats.vospace.datamodel.NodeUtils;
import static it.inaf.oats.vospace.datamodel.NodeUtils.urlEncodePath;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;
import javax.servlet.http.HttpServletRequest;
import net.ivoa.xml.uws.v1.ExecutionPhase;
import net.ivoa.xml.uws.v1.JobSummary;
import net.ivoa.xml.vospace.v2.ContainerNode;
import net.ivoa.xml.vospace.v2.Node;
import net.ivoa.xml.vospace.v2.Param;
import net.ivoa.xml.vospace.v2.Property;
import net.ivoa.xml.vospace.v2.Protocol;
import net.ivoa.xml.vospace.v2.Transfer;
import net.ivoa.xml.vospace.v2.View;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

@RestController
public class NodesController extends BaseController {

    private static final Logger LOG = LoggerFactory.getLogger(NodesController.class);

    @Value("${vospace-authority}")
    private String authority;

    @Value("${maxPollingAttempts:10}")
    private int maxPollingAttempts;

    @Autowired
    private VOSpaceClient client;

    @Autowired
    private HttpServletRequest servletRequest;

    @GetMapping(value = {"/nodes", "/nodes/**"}, produces = MediaType.APPLICATION_JSON_VALUE)
    public ResponseEntity<ListNodeData> listNodes(User principal) throws Exception {

        String path = getPath("/nodes/");
        LOG.debug("listNodes called for path {}", path);

        ListNodeData listNodeData = new ListNodeData();

        Node node = client.getNode(path);

        listNodeData.setWritable(NodeUtils.checkIfWritable(node, principal.getName(), principal.getGroups()));

        MainNodesHtmlGenerator htmlGenerator = new MainNodesHtmlGenerator(node, principal, authority);
        listNodeData.setHtml(htmlGenerator.generateNodes());

        return ResponseEntity.ok(listNodeData);
    }

    @GetMapping(value = "/nodesForMove", produces = MediaType.APPLICATION_JSON_VALUE)
    public ResponseEntity<ListNodeData> listNodesForMoveModal(@RequestParam("path") String path, @RequestParam("nodeToMove") String nodeToMove, User principal) throws Exception {

        LOG.debug("listNodes called for path {}", path);

        ListNodeData listNodeData = new ListNodeData();

        Node node = client.getNode(path);

        listNodeData.setWritable(NodeUtils.checkIfWritable(node, principal.getName(), principal.getGroups()));

        MoveNodeModalHtmlGenerator htmlGenerator = new MoveNodeModalHtmlGenerator(node, nodeToMove, principal, authority);
        listNodeData.setHtml(htmlGenerator.generateNodes());

        return ResponseEntity.ok(listNodeData);
    }

    @PostMapping(value = "/folder")
    public void newFolder(@RequestBody Map<String, Object> params) {

        String parentPath = getRequiredParam(params, "parentPath");
        if (!parentPath.startsWith("/")) {
            parentPath = "/" + parentPath;
        }
        String name = getRequiredParam(params, "name");

        LOG.debug("newFolder called for path {}/{}", parentPath, name);

        ContainerNode node = new ContainerNode();
        node.setUri("vos://" + authority + urlEncodePath(parentPath + "/" + name));

        Property creator = new Property();
        creator.setUri("ivo://ivoa.net/vospace/core#creator");
        creator.setValue(getUser().getName());
        node.getProperties().add(creator);

        client.createNode(node);
    }

    @PostMapping(value = "/delete", consumes = MediaType.APPLICATION_JSON_VALUE)
    public ResponseEntity<?> deleteNodes(@RequestBody List<String> nodesToDelete) {

        CompletableFuture[] deleteCalls = nodesToDelete.stream()
                .map(nodeToDelete -> {
                    return CompletableFuture.runAsync(() -> {
                        LOG.debug("deleteNode called for path {}", nodeToDelete);
                        client.deleteNode(nodeToDelete);
                    }, Runnable::run);
                })
                .collect(Collectors.toList())
                .toArray(CompletableFuture[]::new);

        Optional<RuntimeException> error = CompletableFuture.allOf(deleteCalls)
                .handle((fn, ex) -> {
                    return Optional.ofNullable(ex == null ? null : new RuntimeException(ex));
                }).join();

        if (error.isPresent()) {
            throw error.get();
        }

        // All the nodes have been correctly deleted
        return ResponseEntity.noContent().build();
    }

    @PostMapping(value = "/zip", consumes = MediaType.APPLICATION_JSON_VALUE)
    public Job createZip(@RequestBody List<String> paths) {
        return createArchive(paths, "ivo://ia2.inaf.it/vospace/views#zip");
    }

    @PostMapping(value = "/tar", consumes = MediaType.APPLICATION_JSON_VALUE)
    public Job createTar(@RequestBody List<String> paths) {
        return createArchive(paths, "ivo://ia2.inaf.it/vospace/views#tar");
    }

    private Job createArchive(List<String> paths, String viewUri) {

        Transfer transfer = new Transfer();

        View view = new View();
        view.setUri(viewUri);
        transfer.setView(view);

        if (paths.size() == 1) {
            transfer.setTarget("vos://" + authority + paths.get(0));
        } else {
            String parent = getCommonParent(paths);
            transfer.setTarget("vos://" + authority + parent);
            for (String path : paths) {
                String childName = path.substring(parent.length() + 1);
                Param param = new Param();
                param.setUri(viewUri + "/include");
                param.setValue(childName);
                view.getParam().add(param);
            }
        }

        transfer.setDirection("pullFromVoSpace");
        Protocol protocol = new Protocol();
        protocol.setUri("ivo://ivoa.net/vospace/core#httpget");
        transfer.getProtocols().add(protocol);

        return new Job(client.startTransferJob(transfer), Job.JobType.ARCHIVE);
    }

    private String getCommonParent(List<String> vosPaths) {
        String commonParent = null;
        for (String vosPath : vosPaths) {
            if (commonParent == null) {
                commonParent = vosPath;
            } else {
                StringBuilder newCommonParent = new StringBuilder();
                boolean same = true;
                int lastSlashPos = vosPath.lastIndexOf("/");
                for (int i = 0; same && i < Math.min(commonParent.length(), vosPath.length()) && i < lastSlashPos; i++) {
                    if (commonParent.charAt(i) == vosPath.charAt(i)) {
                        newCommonParent.append(commonParent.charAt(i));
                    } else {
                        same = false;
                    }
                }
                commonParent = newCommonParent.toString();
            }
        }
        return commonParent;
    }

    @PostMapping(value = "/move")
    public ResponseEntity<Job> moveNode(@RequestBody Map<String, Object> params) {

        String target = urlEncodePath(getRequiredParam(params, "target"));
        String direction = urlEncodePath(getRequiredParam(params, "direction"));

        Transfer transfer = new Transfer();
        transfer.setTarget("vos://" + authority + target);
        transfer.setDirection("vos://" + authority + direction);

        JobSummary job = client.startTransferJob(transfer);

        // Try to perform polling until completion. If it takes too much time 
        // sends the execution phase to the UI and let it handles the polling.
        ExecutionPhase phase;
        int i = 0;
        do {
            phase = client.getJobPhase(job.getJobId());
            if (phase == ExecutionPhase.COMPLETED) {
                break;
            } else if (phase == ExecutionPhase.ERROR) {
                String errorDetail = client.getErrorDetail(job.getJobId());
                throw new VOSpaceException("MoveNode operation failed: " + errorDetail);
            }
            try {
                Thread.sleep(1000);
            } catch (InterruptedException ex) {
            }
            i++;
        } while (i < maxPollingAttempts);

        job.setPhase(phase);

        return ResponseEntity.ok(new Job(job, Job.JobType.MOVE));
    }

    protected String getPath(String prefix) {
        String requestURL = servletRequest.getRequestURL().toString();
        return NodeUtils.getPathFromRequestURLString(requestURL, prefix);
    }
}
